@@ -84,6 +84,7 @@ struct event_t {
 	u64 ts;
 	u64 print_skb_id;
 	u64 print_shinfo_id;
+	u64 print_bpfmap_id;
 	struct skb_meta meta;
 	struct tuple tuple;
 	struct tuple tunnel_tuple;
@@ -137,6 +138,17 @@ struct {
 	__uint(max_entries, MAX_TRACK_SIZE);
 } xdp_dhs_skb_heads SEC(".maps");
 
+/* fexit/xdp runs unconditionally for all XDP packets, but we only want to
+ * track (for later SKB tracing) those that passed the filter in fentry/xdp.
+ * This per-CPU map indicates whether the current XDP packet matched the filter.
+ */
+struct {
+	__uint(type, BPF_MAP_TYPE_PERCPU_ARRAY);
+	__type(key, u32);
+	__type(value, bool);
+	__uint(max_entries, 1);
+} xdp_filter_matched SEC(".maps");
+
 struct config {
 	u32 netns;
 	u32 mark;
@@ -173,6 +185,13 @@ struct {
 	__uint(value_size, MAX_STACK_DEPTH * sizeof(u64));
 } print_stack_map SEC(".maps");
 
+struct {
+	__uint(type, BPF_MAP_TYPE_PERCPU_ARRAY);
+	__uint(max_entries, 1);
+	__type(key, u32);
+	__type(value, struct event_t);
+} event_stash SEC(".maps");
+
 struct print_skb_value {
 	u32 len;
 	char str[PRINT_SKB_STR_SIZE];
@@ -181,6 +200,14 @@ struct print_shinfo_value {
 	u32 len;
 	char str[PRINT_SHINFO_STR_SIZE];
 };
+struct print_bpfmap_value {
+	u32 id;
+	char name[16];
+	u32 key_size;
+	u32 value_size;
+	u8 key[256];
+	u8 value[256];
+} __attribute__((packed));
 struct {
 	__uint(type, BPF_MAP_TYPE_PERCPU_ARRAY);
 	__uint(max_entries, 1);
@@ -205,6 +232,18 @@ struct {
 	__type(key, u64);
 	__type(value, struct print_shinfo_value);
 } print_shinfo_map SEC(".maps");
+struct {
+	__uint(type, BPF_MAP_TYPE_PERCPU_ARRAY);
+	__uint(max_entries, 1);
+	__type(key, u32);
+	__type(value, u32);
+} print_bpfmap_id_map SEC(".maps");
+struct {
+	__uint(type, BPF_MAP_TYPE_HASH);
+	__uint(max_entries, 1024);
+	__type(key, u64);
+	__type(value, struct print_bpfmap_value);
+} print_bpfmap_map SEC(".maps");
 
 struct {
 	__uint(type, BPF_MAP_TYPE_PERCPU_ARRAY);
@@ -242,6 +281,12 @@ filter_meta(struct sk_buff *skb) {
 	return true;
 }
 
+static __noinline bool
+filter_skb_expr(struct sk_buff *skb)
+{
+	return skb != NULL;
+}
+
 static __noinline bool
 filter_pcap_ebpf_l3(void *_skb, void *__skb, void *___skb, void *data, void* data_end)
 {
@@ -320,7 +365,7 @@ filter_pcap(struct sk_buff *skb) {
 
 static __always_inline bool
 filter(struct sk_buff *skb) {
-	return filter_pcap(skb) && filter_meta(skb);
+	return filter_pcap(skb) && filter_meta(skb) && filter_skb_expr(skb);
 }
 
 static __noinline void
@@ -414,7 +459,7 @@ static __always_inline u64
 sync_fetch_and_add(void *id_map) {
 	u32 *id = bpf_map_lookup_elem(id_map, &ZERO);
 	if (id)
-		return ((*id)++) | ((u64)bpf_get_smp_processor_id() << 32);
+		return ((*id)++) | ((u64)(bpf_get_smp_processor_id() + 1) << 32);
 	return 0;
 }
 
@@ -484,19 +529,18 @@ get_tracing_fp(void)
 	return fp;
 }
 
-#ifdef bpf_target_arm64
-static __always_inline u64 detect_tramp_fp(void);
+#if defined(bpf_target_arm64) || defined(bpf_target_riscv)
+static __always_inline u64 detect_tramp_fp(void *ctx);
 #endif
 
 static __always_inline u64
-get_tramp_fp(void) {
+get_tramp_fp(void *ctx) {
 	u64 fp_tramp = 0;
 
 #ifdef bpf_target_x86
-	u64 fp = get_tracing_fp();
-	bpf_probe_read_kernel(&fp_tramp, sizeof(fp_tramp), (void *) fp);
-#elif defined(bpf_target_arm64)
-	fp_tramp = detect_tramp_fp();
+	fp_tramp = (u64) ctx + 8;
+#elif defined(bpf_target_arm64) || defined(bpf_target_riscv)
+	fp_tramp = detect_tramp_fp(ctx);
 #endif
 
 	return fp_tramp;
@@ -511,7 +555,7 @@ get_kprobe_fp(struct pt_regs *ctx)
 static __always_inline u64
 get_stackid(void *ctx, const bool is_kprobe) {
 	u64 caller_fp;
-	u64 fp = is_kprobe ? get_kprobe_fp(ctx) : get_tramp_fp();
+	u64 fp = is_kprobe ? get_kprobe_fp(ctx) : get_tramp_fp(ctx);
 	for (int depth = 0; depth < MAX_STACK_DEPTH; depth++) {
 		if (bpf_probe_read_kernel(&caller_fp, sizeof(caller_fp), (void *)fp) < 0)
 			break;
@@ -624,27 +668,25 @@ cont:
 	return true;
 }
 
-static __always_inline u64 get_func_ip(void);
+static __always_inline u64 get_func_ip(void *ctx);
 
 static __always_inline u64
-get_addr(void *ctx, const bool is_kprobe, const bool has_get_func_ip) {
-	u64 ip;
-
-	if (has_get_func_ip) {
-		ip = bpf_get_func_ip(ctx); /* endbr has been handled in helper. */
-	} else {
-		ip = is_kprobe ? PT_REGS_IP((struct pt_regs *) ctx) : get_func_ip();
-#ifdef bpf_target_x86
-		ip -= ENDBR_INSN_SIZE;
-		ip -= is_kprobe; /* -1 always on x86 if kprobe. */
-#endif
+get_addr(void *ctx, const bool is_kprobe) {
+	if (bpf_core_enum_value_exists(enum bpf_func_id, BPF_FUNC_get_func_ip)) {
+		/* endbr has been handled in helper. */
+		return bpf_get_func_ip(ctx);
 	}
 
+	u64 ip = is_kprobe ? PT_REGS_IP((struct pt_regs *) ctx) : get_func_ip(ctx);
+#ifdef bpf_target_x86
+	ip -= ENDBR_INSN_SIZE;
+	ip -= is_kprobe; /* -1 always on x86 if kprobe. */
+#endif
 	return ip;
 }
 
 static __always_inline int
-kprobe_skb(struct sk_buff *skb, struct pt_regs *ctx, const bool has_get_func_ip,
+kprobe_skb(struct sk_buff *skb, struct pt_regs *ctx,
 	   u64 *_stackid, const bool kprobe_multi) {
 	struct event_t event = {};
 
@@ -652,7 +694,7 @@ kprobe_skb(struct sk_buff *skb, struct pt_regs *ctx, const bool has_get_func_ip,
 		return BPF_OK;
 
 	event.skb_addr = (u64) skb;
-	event.addr = get_addr(ctx, true, has_get_func_ip);
+	event.addr = get_addr(ctx, true);
 	event.type = kprobe_multi ? EVENT_TYPE_KPROBE_MULTI: EVENT_TYPE_KPROBE;
 	event.param_second = PT_REGS_PARM2(ctx);
 	event.param_third = PT_REGS_PARM3(ctx);
@@ -669,13 +711,13 @@ kprobe_skb(struct sk_buff *skb, struct pt_regs *ctx, const bool has_get_func_ip,
 SEC("kprobe/skb-" #X)								\
 	int kprobe_skb_##X(struct pt_regs *ctx) {				\
 		struct sk_buff *skb = (struct sk_buff *) PT_REGS_PARM##X(ctx);	\
-		return kprobe_skb(skb, ctx, false, NULL, false);		\
+		return kprobe_skb(skb, ctx, NULL, false);		\
 	}									\
 										\
 	SEC("kprobe.multi/skb-" #X)						\
 	int kprobe_multi_skb_##X(struct pt_regs *ctx) {				\
 		struct sk_buff *skb = (struct sk_buff *) PT_REGS_PARM##X(ctx);	\
-		return kprobe_skb(skb, ctx, true, NULL, true);			\
+		return kprobe_skb(skb, ctx, NULL, true);			\
 	}
 
 PWRU_ADD_KPROBE(1)
@@ -692,7 +734,7 @@ int kprobe_skb_by_stackid(struct pt_regs *ctx) {
 
 	struct sk_buff **skb = bpf_map_lookup_elem(&stackid_skb, &stackid);
 	if (skb && *skb)
-		return kprobe_skb(*skb, ctx, false, &stackid, false);
+		return kprobe_skb(*skb, ctx, &stackid, false);
 
 	return BPF_OK;
 }
@@ -739,43 +781,51 @@ int BPF_PROG(fexit_skb_copy, struct sk_buff *old, gfp_t mask, struct sk_buff *ne
 	return BPF_OK;
 }
 
-#ifdef bpf_target_arm64
-/* As R10 of bpf is not A64_FP, we need to detect the FP of trampoline
- * by scanning the stacks of current bpf prog and the trampoline.
+#if defined(bpf_target_arm64) || defined(bpf_target_riscv)
+/* As R10 of bpf is not A64_FP nor RISC-V fp, we need to detect the FP of
+ * trampoline by scanning the stacks of current bpf prog and the trampoline.
  *
- * Since commit 5d4fa9ec5643 ("bpf, arm64: Avoid blindly saving/restoring all callee-saved registers"),
+ * On arm64, since commit 5d4fa9ec5643 ("bpf, arm64: Avoid blindly saving/restoring all callee-saved registers"),
  * the number of callee-saved registers saved in the bpf prog prologue is
  * dynamic, not fixed anymore.
  */
 static __always_inline u64
-detect_tramp_fp(void) {
-	static const int range_of_detection = 256;
-	u64 fp, r10;
+detect_tramp_fp(void *ctx) {
+	u64 fp, ip, ptr;
 
-	r10 = get_tracing_fp(); /* R10 of current bpf prog */
-	for (int i = 6; i >= 0; i--) {
-		bpf_probe_read_kernel(&fp, sizeof(fp), (void *) (r10 + i * 16));
-		if (r10 < fp && fp < r10 + range_of_detection)
-			return fp;
+	ip = bpf_get_func_ip(ctx);
+	ip += 8; /* The IP of tracee stored on stack has 8B far from its original entry. */
+	ptr = (u64) ctx + 8 /* skb/xdp */ + 16 /* x19 and x20 */ + 8 /* fp */;
+
+	/*
+	 * As we are not sure about the size of `padding`, detect the FP of
+	 * the trampoline by checking the IP of tracee.
+	 */
+	for (int i = 0; i < 6; i++) {
+		bpf_probe_read_kernel(&fp, sizeof(fp), (void *) (ptr + i * 8));
+		if (fp == ip)
+			return fp - 8;
 	}
 
-	return r10;
+	return fp;
 }
 #endif
 
 static __always_inline u64
-get_func_ip(void) {
+get_func_ip(void *ctx) {
 	u64 fp_tramp, ip;
 
 #if defined(bpf_target_x86)
 	static const int ip_offset = 5/* sizeof callq insn */;
-	u64 fp;
 #elif defined(bpf_target_arm64)
 	/* Ref: commit b2ad54e1533e ("bpf, arm64: Implement bpf_arch_text_poke() for arm64") */
 	static const int ip_offset = 12/* sizeof 3 insns */;
 #elif defined(bpf_target_loongarch)
 	/* LoongArch does not support trampoline, get_func_ip does not apply to LoongArch */
 	static const int ip_offset = 0;
+#elif defined(bpf_target_riscv)
+	/* Ref: commit 25ad10658dc1 ("riscv, bpf: Adapt bpf trampoline to optimized riscv ftrace framework") */
+	static const int ip_offset = 8/* sizeof 2 insns */;
 #else
 #error Unsupported architecture
 #endif
@@ -787,6 +837,8 @@ get_func_ip(void) {
 	 * | rip | IP of tracee
 	 * | rbp | FP of tracee's caller
 	 * +-----+ FP of trampoline
+	 * | skb |
+	 * +-----+ ctx of fentry
 	 * | ... |
 	 * | rip | IP of trampoline
 	 * | rbp | FP of trampoline
@@ -797,30 +849,31 @@ get_func_ip(void) {
 	 * Stack layout on arm64:
 	 * |  r9  |
 	 * |  fp  | FP of tracee's caller
+	 * +------+ FP of tracee
 	 * |  lr  | IP of tracee
 	 * |  fp  | FP of tracee
-	 * +------+ FP of trampoline  <-------+
-	 * |  ..  | padding                   |
-	 * |  ..  | callee saved regs         |
-	 * | retv | retval of tracee          |
-	 * | regs | regs of tracee            |
-	 * | nreg | number of regs            |
-	 * |  ip  | IP of tracee if needed    | possible range of
-	 * | rctx | bpf_tramp_run_ctx         | detection
-	 * |  lr  | IP of trampoline          |
-	 * |  fp  | FP of trampoline  <--------- detect it
-	 * +------+ FP of current prog        |
-	 * | regs | callee saved regs         |
-	 * +------+ R10 of bpf prog   <-------+
+	 * +------+ FP of trampoline  <--------- detect it
+	 * |  ..  | padding
+	 * |  x20 | always
+	 * |  x19 | always
+	 * |  arg | skb/xdp always for our case
+	 * +------+ ctx of current prog
+	 * | nreg | number of regs
+	 * |  ip  | IP of tracee
+	 * | rctx | bpf_tramp_run_ctx
+	 * |  lr  | IP of trampoline
+	 * |  fp  | FP of trampoline
+	 * +------+ FP of current prog
+	 * | regs | callee saved regs
+	 * +------+ R10 of bpf prog
 	 * |  ..  |
 	 * +------+ SP of current prog
 	 */
 
 #if defined(bpf_target_x86)
-	fp = get_tracing_fp(); /* FP of current prog */
-	bpf_probe_read_kernel(&fp_tramp, sizeof(fp_tramp), (void *)fp); /* FP of trampoline */
-#elif defined(bpf_target_arm64)
-	fp_tramp = detect_tramp_fp(); /* FP of trampoline */
+	fp_tramp = get_tramp_fp(ctx);
+#elif defined(bpf_target_arm64) || defined(bpf_target_riscv)
+	return bpf_get_func_ip(ctx); /* always supported on arm64 */
 #endif
 	bpf_probe_read_kernel(&ip, sizeof(ip), (void *)(fp_tramp + 8)); /* IP of tracee */
 	return ip - ip_offset;
@@ -834,7 +887,7 @@ int BPF_PROG(fentry_tc, struct sk_buff *skb) {
 		return BPF_OK;
 
 	event.skb_addr = (u64) skb;
-	event.addr = get_addr(ctx, false, false);
+	event.addr = get_addr(ctx, false);
 	event.type = EVENT_TYPE_TC;
 	bpf_map_push_elem(&events, &event, BPF_EXIST);
 
@@ -863,6 +916,12 @@ filter_xdp_meta(struct xdp_buff *xdp) {
 	return filter_xdp_netns(xdp) && filter_xdp_ifindex(xdp);
 }
 
+static __noinline bool
+filter_xdp_expr(struct xdp_buff *xdp)
+{
+	return xdp != NULL;
+}
+
 static __always_inline bool
 filter_xdp_pcap(struct xdp_buff *xdp) {
 	void *data = (void *)(long) BPF_CORE_READ(xdp, data);
@@ -872,13 +931,13 @@ filter_xdp_pcap(struct xdp_buff *xdp) {
 
 static __always_inline bool
 filter_xdp(struct xdp_buff *xdp) {
-	return filter_xdp_pcap(xdp) && filter_xdp_meta(xdp);
+	return filter_xdp_pcap(xdp) && filter_xdp_meta(xdp) && filter_xdp_expr(xdp);
 }
 
 static __noinline void
 set_xdp_metadata(struct xdp_buff *xdp, u64 *metadata) {
 	/* This func will be rewrote by Go. */
-	metadata[0] = (u64)(void *) xdp;
+	metadata[0] = xdp ? 1 : 0;
 }
 
 static __always_inline void
@@ -933,14 +992,17 @@ SEC("fentry/xdp")
 int BPF_PROG(fentry_xdp, struct xdp_buff *xdp) {
 	struct event_t event = {};
 	u64 xdp_dhs = (u64) BPF_CORE_READ(xdp, data_hard_start);
+	bool matched = false;
 
 	if (cfg->is_set) {
 		if (cfg->track_skb && bpf_map_lookup_elem(&xdp_dhs_skb_heads, &xdp_dhs)) {
 			bpf_map_delete_elem(&xdp_dhs_skb_heads, &xdp_dhs);
+			matched = true;
 			goto cont;
 		}
 
 		if (filter_xdp(xdp)) {
+			matched = true;
 			goto cont;
 		}
 
@@ -950,11 +1012,15 @@ cont:
 		set_xdp_output(ctx, xdp, &event);
 	}
 
+	if (cfg->track_skb && matched) {
+		bpf_map_update_elem(&xdp_filter_matched, &ZERO, &TRUE, BPF_ANY);
+	}
+
 	event.pid = bpf_get_current_pid_tgid() >> 32;
 	event.ts = bpf_ktime_get_ns();
 	event.cpu_id = bpf_get_smp_processor_id();
 	event.skb_addr = (u64) &xdp;
-	event.addr = get_addr(ctx, false, false);
+	event.addr = get_addr(ctx, false);
 	event.type = EVENT_TYPE_XDP;
 	bpf_map_push_elem(&events, &event, BPF_EXIST);
 
@@ -963,8 +1029,12 @@ cont:
 
 SEC("fexit/xdp")
 int BPF_PROG(fexit_xdp, struct xdp_buff *xdp) {
-	u64 xdp_dhs = (u64) BPF_CORE_READ(xdp, data_hard_start);
-	bpf_map_update_elem(&xdp_dhs_skb_heads, &xdp_dhs, &xdp, BPF_ANY);
+	bool *matched = bpf_map_lookup_elem(&xdp_filter_matched, &ZERO);
+	if (matched && *matched) {
+		u64 xdp_dhs = (u64) BPF_CORE_READ(xdp, data_hard_start);
+		bpf_map_update_elem(&xdp_dhs_skb_heads, &xdp_dhs, &xdp, BPF_ANY);
+		*matched = false;
+	}
 	return BPF_OK;
 }
 
@@ -995,4 +1065,114 @@ int kretprobe_veth_convert_skb_to_xdp_buff(struct pt_regs *ctx) {
 	return BPF_OK;
 }
 
+static __always_inline void
+set_common_bpfmap_info(struct pt_regs *ctx, u64 *event_id,
+		       struct print_bpfmap_value *bpfmap) {
+	struct bpf_map *map = (struct bpf_map *)PT_REGS_PARM1(ctx);
+
+	*event_id = sync_fetch_and_add(&print_bpfmap_id_map);
+	BPF_CORE_READ_INTO(&bpfmap->id, map, id);
+	BPF_CORE_READ_STR_INTO(&bpfmap->name, map, name);
+	BPF_CORE_READ_INTO(&bpfmap->key_size, map, key_size);
+	BPF_CORE_READ_INTO(&bpfmap->value_size, map, value_size);
+	bpf_probe_read_kernel(&bpfmap->key, sizeof(bpfmap->key), (void *)PT_REGS_PARM2(ctx));
+}
+
+SEC("kprobe/bpf_map_update_elem")
+int kprobe_bpf_map_update_elem(struct pt_regs *ctx) {
+	u64 stackid = get_stackid(ctx, true);
+
+	struct sk_buff **skb = bpf_map_lookup_elem(&stackid_skb, &stackid);
+	if (skb && *skb) {
+		struct event_t event = {};
+
+		event.addr = get_addr(ctx, true);
+		if (!handle_everything(*skb, ctx, &event, &stackid, true))
+			return BPF_OK;
+
+
+		static struct print_bpfmap_value bpfmap = {};
+		set_common_bpfmap_info(ctx, &event.print_bpfmap_id, &bpfmap);
+		bpf_probe_read_kernel(&bpfmap.value,
+				      sizeof(bpfmap.value),
+				      (void *)PT_REGS_PARM3(ctx));
+
+		bpf_map_update_elem(&print_bpfmap_map, &event.print_bpfmap_id, &bpfmap, BPF_ANY);
+		bpf_map_push_elem(&events, &event, BPF_EXIST);
+	}
+
+	return BPF_OK;
+}
+
+SEC("kprobe/bpf_map_delete_elem")
+int kprobe_bpf_map_delete_elem(struct pt_regs *ctx) {
+	u64 stackid = get_stackid(ctx, true);
+
+	struct sk_buff **skb = bpf_map_lookup_elem(&stackid_skb, &stackid);
+	if (skb && *skb) {
+		struct event_t event = {};
+
+		event.addr = get_addr(ctx, true);
+		if (!handle_everything(*skb, ctx, &event, &stackid, true))
+			return BPF_OK;
+
+		static struct print_bpfmap_value bpfmap = {};
+		set_common_bpfmap_info(ctx, &event.print_bpfmap_id, &bpfmap);
+
+		bpf_map_update_elem(&print_bpfmap_map, &event.print_bpfmap_id, &bpfmap, BPF_ANY);
+		bpf_map_push_elem(&events, &event, BPF_EXIST);
+	}
+
+	return BPF_OK;
+}
+
+SEC("kprobe/bpf_map_lookup_elem")
+int kprobe_bpf_map_lookup_elem(struct pt_regs *ctx) {
+	u64 stackid = get_stackid(ctx, true);
+
+	struct sk_buff **skb = bpf_map_lookup_elem(&stackid_skb, &stackid);
+	if (skb && *skb) {
+		struct event_t event = {};
+
+		event.addr = get_addr(ctx, true);
+		if (!handle_everything(*skb, ctx, &event, &stackid, true))
+			return BPF_OK;
+
+		static struct print_bpfmap_value bpfmap = {};
+		set_common_bpfmap_info(ctx, &event.print_bpfmap_id, &bpfmap);
+
+		bpf_map_update_elem(&print_bpfmap_map, &event.print_bpfmap_id, &bpfmap, BPF_ANY);
+		bpf_map_update_elem(&event_stash, &ZERO, &event, BPF_ANY);
+	}
+
+	return BPF_OK;
+}
+
+SEC("kretprobe/bpf_map_lookup_elem")
+int kretprobe_bpf_map_lookup_elem(struct pt_regs *ctx) {
+	/* Two assumptions:
+	 * 1. CPU won't be preempted between kprobe and kretprobe of the same
+	 * lookup operation.
+	 * 2. Lookup won't be recursive.
+	 *
+	 * I believe both are true in the current implementation of BPF.
+	 * Therefore, using PERCPU array to stash the event is safe.
+	 */
+	struct event_t *event = bpf_map_lookup_elem(&event_stash, &ZERO);
+	if (!event)
+		return BPF_OK;
+
+	struct print_bpfmap_value *bpfmap = bpf_map_lookup_elem(&print_bpfmap_map,
+								   &event->print_bpfmap_id);
+	if (!bpfmap)
+		return BPF_OK;
+
+	bpf_probe_read_kernel(&bpfmap->value,
+			      sizeof(bpfmap->value),
+			      (void *)PT_REGS_RC(ctx));
+
+	bpf_map_push_elem(&events, event, BPF_EXIST);
+	return BPF_OK;
+}
+
 char __license[] SEC("license") = "Dual BSD/GPL";
